import numba, numpy as np , math 


###################################################
# Fortran conversions
###################################################
@numba.njit
def sign(a,b) : 
    if b >= 0.0 : return abs(a)
    return -abs(a)

    

###################################################
# Brent minimisation
###################################################
@numba.njit
def brent(x1,x2,  e, incl, w, radius_1):
    # pars
    tol = 1e-8
    itmax = 100
    eps = 1e-5

    a = x1
    b = x2
    c = 0.
    d = 0.
    e = 0.
    fa = get_z(a, e, incl, w, radius_1)
    fb = get_z(b, e, incl, w, radius_1)

    fc = fb

    for iter in range(itmax):
        if (fb*fc > 0.0):
            c = a
            fc = fa
            d = b-a
            e=d   

        if (abs(fc) < abs(fb)):
            a = b
            b = c
            c = a
            fa = fb
            fb = fc
            fc = fa

        tol1 = 2.0*eps*abs(b)+0.5*tol
        xm = (c-b)/2.0
        if (abs(xm) <  tol1 or fb == 0.0) : return b

        if (abs(e) > tol1 and abs(fa) >  abs(fb)):
            s = fb/fa
            if (a == c):
                p = 2.0*xm*s
                q = 1.0-s
            else:
                q = fa/fc
                r = fb/fc
                p = s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0))
                q = (q-1.0)*(r-1.0)*(s-1.0)
            
            if (p > 0.0) : q = - q
            p = abs(p)
            if (2.0*p < min(3.0*xm*q-abs(tol1*q),abs(e*q))):
                e = d
                d = p/q
            else:
                d = xm
                e = d
        else:
            d = xm
            e = d   

        a = b
        fa = fb      
         
        if( abs(d) > tol1) : b = b + d
        else : b = b + sign(tol1, xm)

        fb = get_z(b, e, incl, w, radius_1)
    return 1

# Gravatational constant 
G = 6.67408e-11

#######################################
#          Keplers quation            #
#######################################

@numba.njit(fastmath=True)
def kepler(M, E, e) : return M - E - e*math.sin(E) 

@numba.njit(fastmath=True)
def kdepler(E, e) : return -1 + e*math.cos(E) 


@numba.njit(fastmath=True)
def fmod(a, b) : return a % abs(b)

#######################################
#        Eccentric anomaly            #
#######################################
@numba.njit(fastmath=True)
def getEccentricAnomaly(M, e, E_tol):
    m =  fmod(M , 2*math.pi) # should be fmod
    f1 = fmod(m , (2*math.pi) + e*math.sin(m)) + e*e*math.sin(2.0*m)/2.0  # should be fmod around 
    test  =1.0 
    e1 = 0.
    e0 = 1.0 
    while (test > E_tol):
        e0 = e1 
        e1 = e0 + (m-(e0 - e*math.sin(e0)))/(1.0 - e*math.cos(e0))
        test = abs(e1-e0)

    if (e1 < 0) : e1 = e1 + 2*math.pi 
    return e1 
 
#######################################
#      Time of periastron passage     #
#######################################
@numba.njit(fastmath=True)
def t_ecl_to_peri(t_ecl, e, w, incl, p_sid,radius_1):
    
    # Define variables used
    efac  = 1.0 - e*2       # check if 2 or power
    sin2i = math.sin(incl)**2.

    # Value of theta for i=90 degrees
    ee = 0.
    theta_0 = (math.pi/2.) - w       # True anomaly at superior conjunction

    # Accurate time of eclipse prior to periastro 
    if incl != math.pi/2 :
        theta = brent(theta_0 - math.pi/2,theta_0 + math.pi/2,  e, incl, w, radius_1)
    else:
        theta = theta_0
    theta = theta_0 # BODGE!

    if (theta == math.pi) :  ee = math.pi
    else :  ee =  2.0 * math.atan(math.sqrt((1.-e)/(1.0+e)) * math.tan(theta/2.0))

    eta = ee - e*math.sin(ee)
    delta_t = eta*p_sid/(2*math.pi)
    return t_ecl  - delta_t 

#######################################
#      Calculate the true anomaly     #
#######################################
@numba.njit(fastmath=True)
def getTrueAnomaly(time, e, w, period,  t_zero, incl, E_tol, radius_1 ):

    # Calcualte the mean anomaly
    M = 2*math.pi*((time -  t_ecl_to_peri(t_zero, e, w, incl, period, radius_1)  )/period % 1.)

    # Calculate the eccentric anomaly
    E = getEccentricAnomaly(M, e, E_tol)

    # Now return the true anomaly
    return 2.*math.atan(math.sqrt((1.+e)/(1.-e))*math.tan(E/2.)) 

#######################################
#  Calculate the projected seperaton  #
#######################################
@numba.njit(fastmath=True)
def get_z(nu, e, incl, w, radius_1) : return (1-e**2) * math.sqrt( 1.0 - math.sin(incl)**2  *  math.sin(nu + w)**2) / (1 + e*math.sin(nu)) /radius_1

#@numba.njit(fastmath=True)
#def get_z_(nu, z) : return get_z(nu, z[0], z[1], z[2], z[3]) # for brent

#######################################
#  Calculate the projected seperaton  #
#######################################
@numba.njit(fastmath=True)
def getProjectedPosition(nu, w, incl) : return math.sin(nu + w)*math.sin(incl) 


#######################################
#     Calculate the mass function     #
#######################################
@numba.njit(fastmath=True)
def mass_function_1(e, P, K1) : return (1-e*e) **1.5*P*86400.1* (K1*1000)**3/(2*math.pi*G*1.989e30)

@numba.njit(fastmath=True)
def mass_function_1_(M2, z) : return ((M2*math.sin(z[1]))**3 / ( (z[0] + M2)**2)) - mass_function_1(z[2], z[3], z[4]) 




# Python imports
import numba, numba.cuda, numpy as np , math 

import matplotlib.pyplot as plt 

# bruce imports
from .kepler import getTrueAnomaly


'''
    *Fit parameters*:
     - epsilon - linear limb dark
     - gamma   - Rp/Rs (ratio of planetary and stellar radius)
     - P       - Orbital period [d]
     - T0      - Central transit time
     - i       - Inclination of orbit [rad]
     - Is      - Inclination of stellar rotation axis [rad]
     - Omega   - Angular rotation velocity (star) [rad/s]
     - lambda  - Sky-projected angle between stellar
       rotation axis and normal of orbit plane [rad]
     - a       - Semi major axis [stellar radii]
'''
#@numba.njit(fastmath=True)
def _Xp(nu, lambda_1, incl, a) : 
    result = -math.cos(lambda_1)*math.sin(nu) - math.sin(lambda_1)*math.cos(incl)*math.cos(nu)
    result *= a
    return result

#@numba.njit(fastmath=True)
def _Zp(nu, lambda_1, incl, a) : 
    result = math.sin(lambda_1)*math.sin(nu) - math.cos(lambda_1)*math.cos(incl)*math.cos(nu)
    result *= a
    return result

#@numba.njit(fastmath=True)
def _g(x, e, g, x0):
    result = (1.0-x**2) * math.asin(math.sqrt((g**2-(x-1.0-e)**2)/(1.0-x**2))) + \
       math.sqrt(2.0*(1.0+e)*(x0-x)*(g**2-(x-1.0-e)**2))
    return result


#@numba.njit(fastmath=True)
def _W1( rho, k):
    result = math.sqrt(1.0 - rho**2) - k**2 * (2.0-rho**2)/(8.0*(1.0-rho**2)**(3.0/2.0))
    return result
  
#@numba.njit(fastmath=True)
def _W2( rho, k):
    result =  math.sqrt(1.0 - rho**2) - k**2 * (4.0-3.0*rho**2)/(8.0*(1.0-rho**2)**(3.0/2.0))
    return result

#@numba.njit(fastmath=True)
def _W3( x0, zeta, xc, etap, k):
    result = math.pi/6.0*(1.0-x0)**2*(2.0+x0) + \
      math.pi/2.0*k*(k-zeta) * \
      _g(xc, etap,k, x0) / _g((1.0-k), -k, k, x0) * \
      _W1(1.0-k, k)
    return result 

#@numba.njit(fastmath=True)
def _W4( x0, zeta, xc, etap, k):
    result = math.pi/8.*(1.0-x0)**2*(1.0+x0)**2 + \
      math.pi/2.*k*(k-zeta)*xc * \
      _g(xc, etap, k, x0) / _g((1.0-k), -k, k, x0) * \
      _W2(1.0-k, k)
    return result

#@numba.njit(fastmath=True)
def _rv( time, RV, RV_err, J,
        t_zero, period, 
        K1, fs, fc,
        V0, dV0,
        incl,
        E_tol,
        v_s, v_c, k, radius_1, q, incl_rot, u_lin_ld,
        loglike_switch):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2

    loglike=0.
    for i in range(time.shape[0]):
        nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol, radius_1 ) 
        model = K1*(e*math.cos(w) + math.cos(nu + w)) + V0 + dV0*(time[i] - t_zero) 
        
        if v_s !=0 and v_c != 0 : 
            vsini_1  = v_s**2 + v_c**2
            lambda_1 = math.atan2(v_s,v_c)
            a = 0.019771142*K1*math.sqrt(1-e**2)*(1+1/q)*period/math.sin(incl)
            omega_1 = math.sqrt(vsini_1**2 / 695510**2)
            print(a, omega_1, vsini_1, lambda_1)

            Xp = _Xp(nu, lambda_1, incl, a)
            Zp = _Zp(nu, lambda_1, incl, a) 
            rho = math.sqrt(Xp**2 + Zp**2) 
            etap = math.sqrt(Xp**2 + Zp**2) - 1.0 
            zeta = (2.0*etap + k**2 + etap**2) / (2.0*(1.0+etap)) 
            x0 = 1.0-(k**2 - etap**2)/(2.0*(1.0+etap))
            xc = x0+(zeta-k)/2.0 

            dphase = abs((time[i] - t_zero)/period) 
            dphase = min( dphase-math.floor(dphase), abs(dphase-math.floor(dphase)-1))

            if (rho < (1.0 - k)) and (dphase < 0.25):
                print('Here1')

                W1 = _W1( rho, k)
                W2 = _W2( rho, k)
                model += Xp*omega_1*math.sin(incl_rot)*k**2 * \
                    (1.0 - u_lin_ld*(1.0 - W2)) / \
                    (1.0 - k**2 - u_lin_ld*(1./3. - k**2*(1.0-W1))) 

            if (rho >= 1.-k) and (rho < 1.0 + k) and (dphase < 0.25):
                print('Here2')
                z0 = math.sqrt((k**2 - etap**2)*( (etap+2.0)**2-k**2) ) / (2.0*(1.0+etap))
                W3 = _W3( x0, zeta, xc, etap, k)
                W4 = _W4( x0, zeta, xc, etap, k)

                model += (Xp*omega_1*math.sin(incl_rot)*( \
                    (1.0-u_lin_ld) * (-z0*zeta + k**2*math.acos(zeta/k)) + \
                    (u_lin_ld/(1.0+etap))*W4)) / \
                    (math.pi*(1.-1.0/3.0*u_lin_ld) - (1.0-u_lin_ld) * (math.asin(z0)-(1.+etap)*z0 + \
                    k**2*math.acos(zeta/k)) - u_lin_ld*W3)
            

        if loglike_switch : 
            wt = 1.0 / (RV_err[i]**2 + J**2)
            loglike += -0.5*((RV[i] - model)**2*wt - math.log(wt))
        else : RV[i] = model 

    return loglike






@numba.cuda.jit
def kernel_rv(time, RV, RV_err, J,
        t_zero, period, 
        K1, fs, fc,
        V0, dV0,
        incl,
        E_tol, 
        loglike):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2

    i = numba.cuda.grid(1)

    nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol ) 
    model = K1*(e*math.cos(w) + math.cos(nu + w)) + V0 + dV0*(time[i] - t_zero) 
    wt = 1.0 / (RV_err[i]**2 + J**2)
    loglike[i] = -0.5*((RV[i] - model)**2*wt - math.log(wt))        


@numba.cuda.reduce
def sum_reduce(a, b):
    return a + b


def rv( time, RV = np.zeros(1), RV_err=np.zeros(1), J=0,
        t_zero=0., period=1., 
        K1=10., fs=0., fc=0.,
        V0=10., dV0=0.,
        incl=90.,
        E_tol=1e-5, 
        v_s=0., v_c=0., k=0.2, radius_1=0.2, q=0., incl_rot=0., u_lin_ld=0.5,
        gpu=0, loglike=np.zeros(1), blocks = 10, threads_per_block = 512):




    # Convert inclination 
    incl = np.pi * incl/180.
    incl_rot = np.pi * incl_rot/180.

    if not gpu:
        # First, let's see if we need loglike or not!
        if RV_err[0]==0 : loglike_switch = 0
        else            : loglike_switch = 1

        # Now, let's initiase the arrays, if needed
        if not loglike_switch : RV = np.empty_like(time) 

        # Now make the call
        loglike = _rv( time, RV, RV_err, J,
            t_zero, period, 
            K1, fs, fc,
            V0, dV0,
            incl,
            E_tol, 
            v_s, v_c, k, radius_1, q, incl_rot, u_lin_ld,
            loglike_switch) 

        if loglike_switch : return loglike 
        else              : return RV 
    
    '''
    if gpu:
        # Loglike ony supported 
        # assumeing loglike is array

        ## Call the kernel to populate loglike 
        kernel_rv[blocks, threads_per_block](time, RV, RV_err, J,
            t_zero, period, 
            K1, fs, fc,
            V0, dV0,
            incl,
            E_tol, 
            loglike)
        
        # let's synchronise to ensure it's finished
        numba.cuda.synchronize() 

        # Now reduce the loglike array
        return sum_reduce(loglike)
    '''




'''
time = np.linspace(2450000, 2450001, 100000)
RV = rv(time)
RV = np.random.normal(RV, 1) 
RV_err = np.random.uniform(0.5,1.5, RV.shape[0])
J = 0.2 

CPU_loglike = rv(time, RV, RV_err, J=J) 
print('CPU loglike = ', CPU_loglike)

# Now copy stuff over to GPU for compliance 
d_time = numba.cuda.to_device(time)
d_RV = numba.cuda.to_device(RV)
d_RV_err = numba.cuda.to_device(RV_err)
loglike = np.empty_like(time) 
d_loglike = numba.cuda.to_device(loglike)

threads_per_block=512 
blocks = int(np.ceil(time.shape[0]/threads_per_block))
GPU_loglike = rv(d_time, d_RV, d_RV_err, J=J, gpu=1, loglike=d_loglike, blocks=blocks, threads_per_block=threads_per_block) 
print('GPU loglike = ', GPU_loglike)



# Time it 
timeit GPU_loglike = rv(d_time, d_RV, d_RV_err, J=J, gpu=1, loglike=d_loglike, blocks=blocks, threads_per_block=threads_per_block) 



plt.plot(time,RV)

plt.figure() 
K1 = np.linspace(5,15, 100)
for i in range(len(K1)) : plt.scatter(K1[i] , rv(time, RV, RV_err, J=J, K1=K1[i]))
plt.show()
'''


import numba, math

@numba.njit
def eker_spots(l, b, i, r, a, u1, u2, phi):
    '''
    !
    ! Analytic light curve for a rotating star with a single circular spot 
    !
    ! Implementation of formulae from Eker, Z., 1994ApJ...420..373E with 
    ! correction for equation for K_3 from erratum 1994ApJ...430..438E.
    !
    ! Input values
    !  l        = longitude of spot centre (radians)
    !  b        = latitude of spot centre (radians)
    !  i        = inclination of rotation axis (radians)
    !  r        = angular radius of spot (radians)
    !  a        = Spot contrast ratio (a=Is/Ip).
    !  u1       = linear limb darkening coefficient.
    !  u2       = quadratic limb darkening coefficient.
    !  phi(nphi) = array of rotation phase values in radians, i.e. twopi*(t-t_0)/P
    !  nphi      = number of phase values
    !
    ! N.B. Eker uses a limb darkening law of the form I_0[1 - u_1.mu + u_2.mu^2], 
    ! i.e., u_2 is the negative of the normal quadratic limb darkening coefficient.
    !
    ! Output values
    !  df(nphi)  = light curve
    !  ii(nphi)  = spot position flag
    !  ifail    = status flag
    !
    ! Return values in the array df are (Fp + Fs)/F where
    !  Fp is the flux from the unspotted photosphere
    !  Fs is the flux from the spot
    !  F is the flux from the star without a spot.
    ! 
    ! Spot position flag ii() is as follows
    !   0 = spot not visible
    !   1 = Spot is on the limb and less than half the spot is visible
    !   2 = Spot is on the limb and more than half the spot is visible.
    !   3 = Spot is completely visible
    !
    ! Return value of ifail is the sum of the following flags.
    !   0 => all ok
    !   1 => r >= pi/2
    !   2 => a < 0.0
    !   4 => nphi < 1
    !
    '''
    pi = math.pi
    halfpi = math.pi/2. 

    # First check
    if (r >= halfpi) : return -99.
    if (a < 0.0)     : return -99.

    # constants 
    cosisinb = math.cos(i)*math.sin(b)
    sinicosb = math.sin(i)*math.cos(b)
    cosr = math.cos(r)
    cosr3 = math.pow(cosr,3)
    cosr4 = math.pow(cosr,4)
    sin2r = math.sin(2.0*r)
    cos2r = math.cos(2.0*r)
    sinr2 = 1.0 - math.pow(cosr,2)
    sinr3  = math.pow(math.sin(r),3)
    sinr4  = math.pow(sinr2,2)
    f0 = (a-1.0)/(1.0-u1/3.0+u2/6.0)
    fn = (1.0-u1+u2)*f0
    fl = (u1-2.0*u2)*f0
    fq = u2*f0
    t13b1 = 2.0*(1.0-cosr3)/3.0
    t13b2 = cosr*sinr2
    t13c1 = 0.5*(1.0-cosr4)
    t13c2 = 0.75*sinr4

    costh0 = cosisinb+sinicosb*math.cos(phi-l)
    th0 = math.acos(costh0)
    sinth02 = 1.0 - math.pow(costh0,2)
    sinth0 = math.sqrt(sinth02)
    tanth0 = sinth0/costh0

    # First, check to make sure the spot isn't around the back 
    if (th0 >= (halfpi+r)) : return 0.
    else:
        # Here, the spot is partly or fully visible
        if (u1 == 0.0) and (u2 == 0.0):
            qn = sinr2*costh0
            ql = 0.0
            qq = 0.0
        
        elif (u2==0.):
            qn = sinr2*costh0
            ql = t13b1 - t13b2*sinth02
            qq = 0.0          
        else:
            qn = sinr2*costh0
            ql = t13b1 - t13b2*sinth02
            qq = t13c1*math.pow(costh0,3) + t13c2*costh0*sinth02

        if (th0 > halfpi):
        
            cosphi0 = -1.0/(tanth0*math.tan(r))
            phi0 = math.acos(cosphi0)
            sinphi0 = math.sin(phi0)
            qn = (phi0*qn - math.asin(cosr/sinth0) - 0.5*sinth0*sinphi0*sin2r)/pi + 0.5
            
            if (u1 != 0.0) or (u2 != 0.0):            
                r0 = abs(th0-halfpi)
                sinr0 = math.sin(r0)
                cosr0 = math.cos(r0)
                sin2r0 = math.sin(2.0*r0)
                cos2r0 = math.cos(2.0*r0)
                # (19a)
                ql = (phi0/3.0*(cosr3-math.pow(cosr0,3))* (1.0-3.0*math.pow(costh0,2)) - (phi0 + sinphi0*cosphi0)*(cosr - cosr0)*sinth02 + 4.0/3.0*sinphi0*(sinr3-math.pow(sinr0,3))*sinth0*costh0 + sinphi0*cosphi0/3.0*(cosr3 - math.pow(cosr0,3))*sinth02)/pi
                if (u2 != 0.0): 
                    K1 = 0.25*phi0*(math.pow(cosr0,4)-cosr4)
                    K2 = -0.125*sinphi0*(r0 - r + 0.5*(sin2r*cos2r-sin2r0*cos2r0))
                    K3 = 0.125*(phi0+sinphi0*cosphi0)*(sinr4-math.pow(sinr0,4))
                    K4 = (sinphi0-math.pow(sinphi0,3)/3.0)*(0.375*(r-r0) + 0.0625*(sin2r*(cos2r-4.0)-sin2r0*(cos2r0-4.0)))
                    qq = (2.0*math.pow(costh0,3)*K1 + 6.0*costh0*sinth0* (costh0*K2 + sinth0*K3) +  2.0*math.pow(sinth0,3)*K4 )/pi     
                  
            
        
        elif (th0 > (halfpi-r)): # Spot is on the limb and more than half the spot is visible..
        
            cosphi0 = -1.0/(tanth0*math.tan(r))
            phi0 = math.acos(cosphi0)
            sinphi0 = math.sin(phi0)
            qn = (phi0*qn - math.asin(cosr/sinth0) - 0.5*sinth0*sinphi0*sin2r)/pi + 0.5

            if ((u1 != 0.0) or (u2 != 0.0)):
                r0 = abs(th0-halfpi)
                sinr0 = math.sin(r0)
                cosr0 = math.cos(r0)
                sin2r0 = math.sin(2.0*r0)
                cos2r0 = math.cos(2.0*r0)

                # (18a)
                Cl = (pi-phi0)/3.0*(cosr3-math.pow(cosr0,3))*(1.0-3.0*math.pow(costh0,2)) - (pi-phi0-sinphi0*cosphi0)*(cosr-cosr0)*sinth02 - 4.0/3.0*sinphi0*(sinr3-math.pow(sinr0,3))*sinth0*costh0 - sinphi0*cosphi0*(cosr3-math.pow(cosr0,3))*sinth02/3.0
                ql = ql - Cl/pi
                if (u2 != 0.0):
                    K1 = 0.25*(pi-phi0)*(math.pow(cosr0,4)-cosr4)
                    K2 = 0.125*sinphi0*(r0 - r + 0.5*(sin2r*cos2r-sin2r0*cos2r0))
                    K3 = 0.125*(pi-phi0+sinphi0*cosphi0)*(sinr4-math.pow(sinr0,4))
                    K4 = -(sinphi0-math.pow(sinphi0,3)/3.0)*(0.375*(r-r0) + 0.0625*(sin2r*(cos2r-4.0)-sin2r0*(cos2r0-4.0)))

                    # (18b)
                    Cq = 2.0*math.pow(costh0,3)*K1 + 6.0*costh0*sinth0*(costh0*K2 + sinth0*K3) + 2.0*math.pow(sinth0,3)*K4
                    qq = qq - Cq/pi
                
        return fn*qn + fl*ql + fq*qq
    


'''
l = 0.2
b = 0.1 
i = math.pi/2.
r = 0.2
a = 0.98 

u1 = 0.5
u2 = 0.2 
import numpy as np 
phis = np.linspace(0,2*math.pi,1000) 

import matplotlib.pyplot as plt 
for i in phis : plt.scatter(i, eker_spots(l, b, i, r, a, u1, u2, i), c='k', s=10) 
plt.show()
'''


import numpy as np 
import matplotlib.pyplot as plt 
import math

from bruce.binarystar.kepler import getTrueAnomaly, get_z
from bruce.binarystar.qpower2 import Flux_drop_analytical_power_2

def unsold(x, epsilon= 0.6 ):
    beta = epsilon/(1-epsilon)
    xc = np.clip(x, -1,1)
    return (2*np.sqrt(1-xc**2) / np.pi + 0.5*beta*(1-xc**2) )/ ( 1 + 2*beta/3)

def area(d,x,R):
    if (x + R < d) : return 0.
    arg1 = (d*d + x*x - R*R)/(2.*d*x)
    arg2 = (d*d + R*R - x*x)/(2.*d*R)
    arg3 = max((-d + x + R)*(d + x - R)*(d - x + R)*(d + x + R), 0.)
    if(x <= R - d)      :  return math.pi*x*x;							# planet completely overlaps stellar circle
    elif(x >= R + d)    :  return math.pi*R*R;						    # stellar circle completely overlaps planet
    else                :  return x*x*math.acos(arg1) + R*R*math.acos(arg2) - 0.5*math.sqrt(arg3) #partial overlap

time = np.linspace(-0.1,0.1,1000)

e = 0.
w = 90.*math.pi/180 
period = 1 
t_zero = 0. 
E_tol = 1e-5 
radius_1 = 0.2
b = 0.9
k = 0.05
incl = math.acos(b*radius_1)
incl_rot = 0.2

Vsini = 10

for i in range(time.shape[0]) : 
    nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol, radius_1 )
    z = get_z(nu, e, incl, w, radius_1) 

    # Model = 
    #   Vsini * Area   *  I / I0
    A = area(z, 1,k)
    r = math.sqrt(np.clip(z**2 - b**2, 0, np.inf)) 
    Vsini_ =   unsold(r, epsilon= 0.6 ) / unsold(0, epsilon= 0.6 )
    Vsini_ = 1-Vsini_ 
    if nu < 0 : Vsini_=-Vsini_ 

    I = Flux_drop_analytical_power_2(z, k, 0.8,0.8, 1e-8)
    I0 = Flux_drop_analytical_power_2(0, k, 0.8,0.8, 1e-8)


    print(time[i], I)
    plt.scatter(time[i], math.cos(b**2)*A*Vsini*Vsini_*I/I0, c='k', s=10)
    plt.scatter(time[i], A*Vsini*Vsini_*I/I0, c='r', s=10)

plt.show()
#getTrueAnomaly(time, e, w, period,  t_zero, incl, E_tol, radius_1 )
#get_z(nu, e, incl, w, radius_1)


import math, numba 

@numba.njit
def clip(a, b, c):
    if (a < b)      :  return b
    elif (a > c)    :  return c
    else            :  return a
@numba.njit
def q1(z, p, c, a, g, I_0):
    zt = clip(abs(z), 0,1-p)
    s = 1-zt*zt
    c0 = (1-c+c*math.pow(s,g))
    c2 = 0.5*a*c*math.pow(s,(g-2))*((a-1)*zt*zt-1)
    return -I_0*math.pi*p*p*(c0 + 0.25*p*p*c2 - 0.125*a*c*p*p*math.pow(s,(g-1)))

@numba.njit
def q2(z, p, c, a, g, I_0, eps):
    zt = clip(abs(z), 1-p,1+p)
    d = clip((zt*zt - p*p + 1)/(2*zt),0,1)
    ra = 0.5*(zt-p+d)
    rb = 0.5*(1+d)
    sa = clip(1-ra*ra,eps,1)
    sb = clip(1-rb*rb,eps,1)
    q = clip((zt-d)/p,-1,1)
    w2 = p*p-(d-zt)*(d-zt)
    w = math.sqrt(clip(w2,eps,1))
    c0 = 1 - c + c*math.pow(sa,g)
    c1 = -a*c*ra*math.pow(sa,(g-1))
    c2 = 0.5*a*c*math.pow(sa,(g-2))*((a-1)*ra*ra-1)
    a0 = c0 + c1*(zt-ra) + c2*(zt-ra)*(zt-ra)
    a1 = c1+2*c2*(zt-ra)
    aq = math.acos(q)
    J1 =  (a0*(d-zt)-(2./3.)*a1*w2 + 0.25*c2*(d-zt)*(2.0*(d-zt)*(d-zt)-p*p))*w + (a0*p*p + 0.25*c2*math.pow(p,4))*aq 
    J2 = a*c*math.pow(sa,(g-1))*math.pow(p,4)*(0.125*aq + (1./12.)*q*(q*q-2.5)*math.sqrt(clip(1-q*q,0.0,1.0)) )
    d0 = 1 - c + c*math.pow(sb,g)
    d1 = -a*c*rb*math.pow(sb,(g-1))
    K1 = (d0-rb*d1)*math.acos(d) + ((rb*d+(2./3.)*(1-d*d))*d1 - d*d0)*math.sqrt(clip(1-d*d,0.0,1.0))
    K2 = (1/3)*c*a*math.pow(sb,(g+0.5))*(1-d)
    if (J1 > 1)   : J1 = 0
    FF =  I_0*(J1 - J2 + K1 - K2)
    #if (FF < 0.9) : FF=1.0
    return -FF


@numba.njit
def Flux_drop_analytical_power_2(z, k, c, a, eps):
    '''
    Calculate the analytical flux drop por the power-2 law.

    Parameters
    z : double
        Projected seperation of centers in units of stellar radii.
    k : double
        Ratio of the radii.
    c : double
        The first power-2 coefficient.
    a : double
        The second power-2 coefficient.
    f : double
        The flux from which to drop light from.
    eps : double
        Factor (1e-9)
    '''
    I_0 = (a+2)/(math.pi*(a-c*a+2))
    g = 0.5*a

    if (z < 1-k)           : return q1(z, k, c, a, g, I_0)
    elif (abs(z-1) < k)    : return q2(z, k, c, a, g, I_0, eps)
    else                   : return 0. 


'''
import numpy as np 
import matplotlib.pyplot as plt 
z = np.linspace(-2,2,1000)
z_ = np.abs(z) 

k = 0.1 
c = 0.8 
a = 0.8 
eps = 1e-8  

for i in range(len(z)) : plt.scatter(z[i], Flux_drop_analytical_power_2(z_[i], k, c, a, eps), c='k', s=10)
plt.show() 
'''


from setuptools import setup, Extension

setup(
    name = 'bruce-package',
    version = '0.9',
    description = 'GPU-accelerated binary star model',
    url = None,
    author = 'Samuel Gill et al',
    author_email = 'samuel.gill@warwick.ac.uk',
    license = 'GNU',
    packages=['bruce','bruce/binarystar'],
    scripts=['Utils/ngtsfit/ngtsfit',
             'Utils/lcbin/lcbin',
             'Utils/tessget/tessget',
             'Utils/tls/tls',
             'Utils/prewhiten/prewhiten',
             'Utils/mbls/mbls',
             'Utils/predict/predict',
             'Utils/analyse/analyse',
             'Utils/analyse/plot_nights',
             'Utils/massfunc/massfunc',
             'Utils/modulation/modulation'],
    install_requires=['celerite', 'numba', 'numpy', 'emcee']
)


# Standard imports
import numba, numba.cuda, numpy as np , math 
from numba import prange 

# bruce imports
from .kepler import getTrueAnomaly, get_z, getProjectedPosition
from .eker_spots import eker_spots
from .doppler import Fdoppler
from .ellipsoidal import Fellipsoidal 
from .quadratic import Flux_drop_analytical_quadratic 
from .qpower2 import Flux_drop_analytical_power_2
from .uniform import Flux_drop_uniform 
from time import time as _time

@numba.njit
def _lc(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, k ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        loglike_switch ):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2 

    nspots = spots.shape[0]//4 

    loglike=0.
    for i in range(time.shape[0]):
        # Calculate the true anomaly 
        nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol, radius_1 ) 

        # Calculate the projected seperation 
        z = get_z(nu, e, incl, w, radius_1)  

        # Initialse the flux
        # The model we will use is:
        #   F_tot = continuum + F_ecl + F_ellipsoidal + F_spots + F_transit 
        #   F_ellipsoidal -> ellipsoidal effect contralled by the mass ratio, q and the gravity limb-darkening coefficient
        #   F_spots       -> flux drop from spots controlled by the eker model
        #   F_transit     -> flux from the desired transit model (using 1 as the continuum)

        continuum = 1.0 # continuum at 1.0 (normalised)
        F_spots = 0.    # spot flux 
        F_doppler = 0.  # Doppler beaming
        F_ellipsoidal = 0.  # Ellipsoidal variation
        F_transit =  0.0    # transit flux drop

        # First, let's check for spots 
        if (nspots > 0):
            spot_phase = omega_1*2*math.pi*(time[i] - t_zero)/period
            for j in range(nspots) : F_spots += eker_spots(spots[j*4 + 0], spots[j*4 +1], incl, spots[j*4 +2], spots[j*4 +3], 0.5,0.3, spot_phase);


        # Next, we need to check for doppler beaming
        # Check for doppler beaming 
        if (alpha_doppler > 0) and (K1 > 0) : F_doppler = Fdoppler(nu, alpha_doppler, K1 )


        # Check for eelipsoidal variation and apply it if needed
        if (q>0.):
            alpha = math.acos(math.sin(w + nu)*math.sin(incl))
            F_ellipsoidal = Fellipsoidal(alpha, q, radius_1, incl, 0.5, gdc_1)

        # Check distance between them to see if its transiting
        if (z < (1.0+ k)):
            # So it's eclipsing, lets find out if its a primary or secondary
            f = getProjectedPosition(nu, w, incl);

            if (f > 0):
                # First, get the flux drop
                if (ld_law_1==0) : F_transit = Flux_drop_uniform( z, k, 1.) # uniform limb-darkening 
                if (ld_law_1==1) : F_transit = Flux_drop_analytical_quadratic( z,  k,  ldc_1_1,  ldc_1_2,  1e-8)
                if (ld_law_1==2) : F_transit = Flux_drop_analytical_power_2(z, k, ldc_1_1, ldc_1_2, 1E-8) 
                if (ld_law_1==-1) : F_transit = -1.

                # Now account for SBR (third light)
                #F_transit = F_transit/(1. + k*k*SBR) + (1.-1.0/(1 + k*k*SBR))

            elif (SBR>0.) :
                    if (ld_law_1!=-1) : F_transit =  Flux_drop_uniform(z, k, SBR) # Secondary eclipse
                    else : F_transit = -1.

        # Now put the model together
        model = continuum + F_spots + F_doppler + F_ellipsoidal + F_transit 

        # That's all from the star, so let's account for third light 
        if (light_3 > 0.0) : model = model/(1. + light_3) + (1.-1.0/(1. + light_3)) # third light


        if loglike_switch : 
            model = zp - 2.5*math.log10(model)
            wt = 1.0 / (LC_ERR[i]**2 + J**2)
            loglike += -0.5*((LC[i] - model)**2*wt - math.log(wt))
        else : LC[i] = model 

    return loglike 



@numba.njit(parallel=True)
def _lc_prange(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, k ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        loglike_switch ):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2 

    nspots = spots.shape[0]//4 

    loglike=0.
    for i in prange(time.shape[0]):
        # Calculate the true anomaly 
        nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol ) 

        # Calculate the projected seperation 
        z = get_z(nu, e, incl, w, radius_1)  

        # Initialse the flux
        # The model we will use is:
        #   F_tot = continuum + F_ecl + F_ellipsoidal + F_spots + F_transit 
        #   F_ellipsoidal -> ellipsoidal effect contralled by the mass ratio, q and the gravity limb-darkening coefficient
        #   F_spots       -> flux drop from spots controlled by the eker model
        #   F_transit     -> flux from the desired transit model (using 1 as the continuum)

        continuum = 1.0 # continuum at 1.0 (normalised)
        F_spots = 0.    # spot flux 
        F_doppler = 0.  # Doppler beaming
        F_ellipsoidal = 0.  # Ellipsoidal variation
        F_transit =  0.0    # transit flux drop

        # First, let's check for spots 
        if (nspots > 0):
            spot_phase = omega_1*2*math.pi*(time[i] - t_zero)/period
            for j in range(nspots) : F_spots += eker_spots(spots[j*4 + 0], spots[j*4 +1], incl, spots[j*4 +2], spots[j*4 +3], 0.5,0.3, spot_phase);


        # Next, we need to check for doppler beaming
        # Check for doppler beaming 
        if (alpha_doppler > 0) and (K1 > 0) : F_doppler = Fdoppler(nu, alpha_doppler, K1 )


        # Check for eelipsoidal variation and apply it if needed
        if (q>0.):
            alpha = math.acos(math.sin(w + nu)*math.sin(incl))
            F_ellipsoidal = Fellipsoidal(alpha, q, radius_1, incl, 0.5, gdc_1)

        # Check distance between them to see if its transiting
        if (z < (1.0+ k)):
            # So it's eclipsing, lets find out if its a primary or secondary
            f = getProjectedPosition(nu, w, incl);

            if (f > 0):
                # First, get the flux drop
                if (ld_law_1==0) : F_transit = Flux_drop_uniform( z, k, 1.) # uniform limb-darkening 
                if (ld_law_1==1) : F_transit = Flux_drop_analytical_quadratic( z,  k,  ldc_1_1,  ldc_1_2,  1e-8)
                if (ld_law_1==2) : F_transit = Flux_drop_analytical_power_2(z, k, ldc_1_1, ldc_1_2, 1E-8) 

                # Now account for SBR (third light)
                #F_transit = F_transit/(1. + k*k*SBR) + (1.-1.0/(1 + k*k*SBR))

            elif (SBR>0.) :  F_transit =  Flux_drop_uniform(z, k, SBR) # Secondary eclipse
        
        # Now put the model together
        model = continuum + F_spots + F_doppler + F_ellipsoidal + F_transit 

        # That's all from the star, so let's account for third light 
        if (light_3 > 0.0) : model = model/(1. + light_3) + (1.-1.0/(1. + light_3)) # third light


        if loglike_switch : 
            model = zp - 2.5*math.log10(model)
            wt = 1.0 / (LC_ERR[i]**2 + J**2)
            loglike += -0.5*((LC[i] - model)**2*wt - math.log(wt))
        else : LC[i] = model 

    return loglike 



@numba.cuda.jit
def kernel_lc(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, k ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        loglike ):

    # Unpack and convert (assume incl in radians)
    w = math.atan2(fs, fc)
    e = fs**2 + fc**2 

    i = numba.cuda.grid(1)


    nspots = spots.shape[0]//4 

    # Calculate the true anomaly 
    nu = getTrueAnomaly(time[i], e, w, period,  t_zero, incl, E_tol, radius_1 ) 

    # Calculate the projected seperation 
    z = get_z(nu, e, incl, w, radius_1)  

    # Initialse the flux
    # The model we will use is:
    #   F_tot = continuum + F_ecl + F_ellipsoidal + F_spots + F_transit 
    #   F_ellipsoidal -> ellipsoidal effect contralled by the mass ratio, q and the gravity limb-darkening coefficient
    #   F_spots       -> flux drop from spots controlled by the eker model
    #   F_transit     -> flux from the desired transit model (using 1 as the continuum)

    continuum = 1.0 # continuum at 1.0 (normalised)
    F_spots = 0.    # spot flux 
    F_doppler = 0.  # Doppler beaming
    F_ellipsoidal = 0.  # Ellipsoidal variation
    F_transit =  0.0    # transit flux drop

    # First, let's check for spots 
    if (nspots > 0):
        spot_phase = omega_1*2*math.pi*(time[i] - t_zero)/period
        for j in range(nspots) : F_spots += eker_spots(spots[j*4 + 0], spots[j*4 +1], incl, spots[j*4 +2], spots[j*4 +3], 0.5,0.3, spot_phase);


    # Next, we need to check for doppler beaming
    # Check for doppler beaming 
    if (alpha_doppler > 0) and (K1 > 0) : F_doppler = Fdoppler(nu, alpha_doppler, K1 )


    # Check for eelipsoidal variation and apply it if needed
    if (q>0.):
        alpha = math.acos(math.sin(w + nu)*math.sin(incl))
        F_ellipsoidal = Fellipsoidal(alpha, q, radius_1, incl, 0.5, gdc_1)

    # Check distance between them to see if its transiting
    if (z < (1.0+ k)):
        # So it's eclipsing, lets find out if its a primary or secondary
        f = getProjectedPosition(nu, w, incl);

        if (f > 0):
            # First, get the flux drop
            if (ld_law_1==0) : F_transit = Flux_drop_uniform( z, k, 1.) # uniform limb-darkening 
            if (ld_law_1==1) : F_transit = Flux_drop_analytical_quadratic( z,  k,  ldc_1_1,  ldc_1_2,  1e-8)
            if (ld_law_1==2) : F_transit = Flux_drop_analytical_power_2(z, k, ldc_1_1, ldc_1_2, 1E-8) 

            # Now account for SBR (third light)
            #F_transit = F_transit/(1. + k*k*SBR) + (1.-1.0/(1 + k*k*SBR))

        elif (SBR>0.) :  F_transit =  Flux_drop_uniform(z, k, SBR) # Secondary eclipse
    
    # Now put the model together
    model = continuum + F_spots + F_doppler + F_ellipsoidal + F_transit 

    # That's all from the star, so let's account for third light 
    if (light_3 > 0.0) : model = model/(1. + light_3) + (1.-1.0/(1. + light_3)) # third light

    model = zp - 2.5*math.log10(model)

    wt = 1.0 / (LC_ERR[i]**2 + J**2)
    loglike[i] = -0.5*((LC[i] - model)**2*wt - math.log(wt))

@numba.cuda.reduce
def sum_reduce(a, b):
    return a + b

def lc(time, LC=np.zeros(1), LC_ERR=np.zeros(1), J=0., zp=0.,
    t_zero=0., period=1.,
    radius_1=0.2, k=0.2 ,
    fs=0., fc=0., 
    q=0., albedo=0.,
    alpha_doppler=0., K1=0.,
    spots = np.zeros(1), omega_1=1., 
    incl = 90.,
    ld_law_1=1, ldc_1_1=0.5, ldc_1_2=0.2, gdc_1=0.4,
    SBR=0., light_3=0.,
    E_tol=1e-5, parallel=False,
    gpu=0, loglike=np.zeros(1), blocks = 10, threads_per_block = 512):

    incl = np.pi * incl/180.


        
    if not gpu:
        # First, let's see if we need loglike or not!
        if LC_ERR[0]==0 : loglike_switch = 0
        else            : loglike_switch = 1

        # Now, let's initiase the arrays, if needed
        if not loglike_switch : LC = np.empty_like(time) 

        # Now make the call
        if parallel:
            loglike = _lc_prange(time, LC, LC_ERR, J, zp,
                t_zero, period,
                radius_1, k ,
                fs, fc, 
                q, albedo,
                alpha_doppler, K1,
                spots, omega_1, 
                incl,
                ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
                SBR, light_3,
                E_tol,
                loglike_switch )
        else:
            loglike = _lc(time, LC, LC_ERR, J, zp,
                t_zero, period,
                radius_1, k ,
                fs, fc, 
                q, albedo,
                alpha_doppler, K1,
                spots, omega_1, 
                incl,
                ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
                SBR, light_3,
                E_tol,
                loglike_switch )
                

        if loglike_switch : return loglike 
        else              : return LC 

    if gpu:
        # Loglike ony supported 
        # assumeing loglike is array

        ## Call the kernel to populate loglike
        #start = _time() 
        kernel_lc[blocks, threads_per_block](time, LC, LC_ERR, J, zp,
            t_zero, period,
            radius_1, k ,
            fs, fc, 
            q, albedo,
            alpha_doppler, K1,
            spots, omega_1, 
            incl,
            ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
            SBR, light_3,
            E_tol,
            loglike )
        #end = _time() 
        #print('Kernel : {:}'.format(end-start))
        #start = _time() 

        # let's synchronise to ensure it's finished
        #numba.cuda.synchronize() 
        #end = _time() 
        #print('sync : {:}'.format(end-start))

        #start = _time() 
        # Now reduce the loglike array
        #lll =  sum_reduce(loglike)
        #end = _time() 
        #print('reduce : {:}'.format(end-start))
        return sum_reduce(loglike)

'''
TEST 3


import numpy as np 
import matplotlib.pyplot as plt  
from bruce import lc 

t = np.linspace(-0.2, 0.8, 100) 
f = lc(t) 
m = -2.5*np.log10(f) 
me = np.random.uniform(0.5e-3,1.5e-3,t.shape[0])
m = np.random.normal(m, me)  

zps = np.linspace(-1,1,1000)

for i in zps : plt.scatter(i, lc(t,m , me, zp=i))
plt.show()
'''


'''
# TEST 2 
import numpy as np 
import matplotlib.pyplot as plt  

t = np.linspace(-0.2, 0.8, 1e6) 
f = lc(t) 
m = -2.5*np.log10(f) 
me = np.random.uniform(0.5e-3,1.5e-3,t.shape[0])
m = np.random.normal(m, me) 

CPU_loglike = lc(t, m, me) 
print('CPU log : ', CPU_loglike)


d_t = numba.cuda.to_device(t)
d_m = numba.cuda.to_device(m)
d_me = numba.cuda.to_device(me) 
d_spots = numba.cuda.to_device(np.array([]))
d_loglike = numba.cuda.to_device(np.empty_like(t)) 

threads_per_block = 256 
blocks = int(np.ceil(t.shape[0]/threads_per_block))


GPU_loglike = lc(d_t, d_m, d_me, gpu=1, loglike=d_loglike, blocks = blocks, threads_per_block = threads_per_block)
print('GPU log : ', GPU_loglike)



plt.scatter(t, m, c='k', s=10)
plt.gca().invert_yaxis() 
plt.show() 
'''
'''

# TEST 1
import numpy as np 
import matplotlib.pyplot as plt 

time = np.linspace(-0.2, 0.8, 1000) 
LC = np.empty_like(time) 
LC_ERR = np.empty_like(time) 
J = 0.1 
zp = 0. 

t_zero = 0. 
period = 1. 
radius_1 = k = 0.2 
fs=fc = 0. 
q = 0.
albedo = 0. 

alpha_doppler = 0.
K1 = 150 

spots = np.array([0.2, 0.2, 0.2, 0.9,4.2, 0.2, 0.2, 0.9]) 
omega_1 = 1.0 

incl = math.pi/2 

ld_law_1 = 1 
ldc_1_1 = 0.5 
ldc_1_2 = 0.2 
gdc_1 = 0.4

SBR = 0.
light_3 = 0. 

E_tol = 1e-5 
logswitch =0

a = _lc(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, k ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        logswitch )


lC = -2.5*np.log10(LC)
LC_ERR = np.random.uniform(0.5e-3, 1.5e-3, time.shape[0])
LC = np.random.normal(LC, LC_ERR)


a = _lc(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, k ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        1 )


plt.scatter(time, LC, c='k', s=10)
plt.title('Loglike : {:}'.format(   a  ))


plt.figure() 

k = np.linspace(0.05,0.4, 100) 
for i in k : plt.scatter(i, _lc(time, LC, LC_ERR, J, zp,
        t_zero, period,
        radius_1, i ,
        fs, fc, 
        q, albedo,
        alpha_doppler, K1,
        spots, omega_1, 
        incl,
        ld_law_1, ldc_1_1, ldc_1_2, gdc_1,
        SBR, light_3,
        E_tol,
        1 ))
plt.show() 
'''


import numba, math

@numba.njit
def Fdoppler(alpha, alpha_doppler, K1 ):
    Ad = alpha_doppler*K1/299792.458
    return Ad*math.sin(alpha -math.pi/2.) 

'''
import numpy as np 
import matplotlib.pyplot as plt 
alpha = np.linspace(0,2*math.pi,100) 
alpha_doppler = 0.1
K1 = 150 

for i in alpha : plt.scatter(i, Fdoppler(i, alpha_doppler, K1 ) ) 
plt.show()
'''